@@ -9,7 +9,7 @@ import redis
 from anytree import Node, NodeMixin
 
 from .conf import config as cfg, FormBackError, r, back_symbol, home_symbol, NavigationBackError, \
-    NavigationInvalidChoice
+    NavigationInvalidChoice, ImproperlyConfigured
 
 LOG_FORMAT = '%(asctime)s %(levelname)-6s %(funcName)s (on line %(lineno)-4d) : %(message)s'
 logging.basicConfig(level=logging.INFO, format=LOG_FORMAT)
@@ -35,9 +35,13 @@ class ShortCutHandler:
     pass
 
 
+def get_var(msisdn, session_id, var):
+    return r.hget(f'{msisdn}:{session_id}', var)
+
+
 class ListInput:
 
-    def __init__(self, items: List, title: str, key=None, idx=None):
+    def __init__(self, items: List, title: str, key=None, idx=None, extra=None):
         """
         For handling Listable items
 
@@ -52,6 +56,7 @@ class ListInput:
         self.title = title
         self.key = key
         self.idx = idx
+        self.extra = extra
 
     def get_items(self):
         if not isinstance(self.items, list):
@@ -61,22 +66,27 @@ class ListInput:
 
         if isinstance(self.items[0], (str, int, float)):
             rsp = '\n'.join([f'{idx}. {str(item)}' for idx, item in enumerate(self.items, start=1)])
-            return f'CON {self.title}\n{rsp}'
+            menu = f'CON {self.title}\n{rsp}'
 
         elif isinstance(self.items[0], dict):
             rsp = '\n'.join([f'{idx}. {item[self.key]}' for idx, item in enumerate(self.items, start=1)])
-            return f'CON {self.title}\n{rsp}'
+            menu = f'CON {self.title}'
 
         elif isinstance(self.items[0], list) or isinstance(self.items[0], tuple):
             rsp = '\n'.join([f'{idx}. {item[idx]}' for idx, item in enumerate(self.items, start=1)])
-            return f'CON {self.title}\n{rsp}'
+            menu = f'CON {self.title}\n{rsp}'
 
         else:
             raise ValueError(
                 f'self.items should contain items of type str, dict, list or tuple, not {self.items[0].__class__.__name__}')
 
+        xtra = '' if self.extra is None else f'\n{self.extra}'
+        return f'{menu}{xtra}'
+
     def get_item(self, idx):
-        return self.items[idx - 1]
+        if isinstance(idx, int):
+            return self.items[idx - 1]
+        return None
 
     def validate(self, key):
         try:
@@ -108,20 +118,26 @@ class FormFlow:
     def get_step_item(self, step):
         return self.form_questions[str(step)]['menu']
 
-    def _validate_last_input(self, current_step, last_input):
+    def call_post_validation(self):
+        pass
+
+    def _validate_last_input(self, current_step, last_input, msisdn, session_id):
         """
         validate by using if...else, for all steps in this flow
         :return:
         """
 
-        _val, _validated = self.step_validator(current_step, last_input)
+        _val, _extra_data = self.step_validator(current_step, last_input, msisdn=msisdn, session_id=session_id)
         if _val is None or not isinstance(_val, bool):
             logger.warning('Input not validated explicitly by validator function, Default value of True has been used')
             _val = True
 
-        return _val
+        return _val, _extra_data
 
-    def _response(self, current_step, last_input):
+    def gather_form_keys(self):
+        return [self.form_questions[x]['name'] for x in self.form_questions.keys()]
+
+    def _response(self, current_step, last_input, msisdn, session_id, ussd_string):
         skip_validation = False
         valid_last_input = False
 
@@ -139,8 +155,28 @@ class FormFlow:
             if self.get_step_type(current_step) == ListInput:
                 list_ref: ListInput = self.get_step_item(current_step)
                 valid_last_input = list_ref.validate(last_input)
+
+                # handle bs logic
+                _res = self._validate_last_input(
+                    current_step, last_input, msisdn=msisdn, session_id=session_id)
+
+                if isinstance(_res, tuple) and len(_res) == 2:
+                    _xtra_data = _res[1]
+                elif isinstance(_res, dict):
+                    _xtra_data = _res
+                else:
+                    raise ImproperlyConfigured(
+                        f"response from {self._validate_last_input}() should be a tuple of bool and dict or a dict."
+                        f" Not {_res.__class__.__name__}")
             else:
-                valid_last_input = self._validate_last_input(current_step, last_input)
+                valid_last_input, _xtra_data = self._validate_last_input(
+                    current_step, last_input, msisdn=msisdn, session_id=session_id)
+
+            if _xtra_data is not None:
+                if isinstance(_xtra_data, dict):
+                    _state.update(_xtra_data)
+                else:
+                    logger.warning(f'extra_data from validation should be a dict not {_xtra_data.__class__.__name__}')
 
         # if last input is valid, display next menu, otherwise, show invalid input message, and display same menu
         if valid_last_input:
@@ -149,14 +185,28 @@ class FormFlow:
                 _field_name: str = self.form_questions.get(str(current_step)).get('name')
                 if _field_name and _field_name.replace("_", "").isalnum() and not _field_name[0].isnumeric():
                     if self.get_step_type(current_step) == ListInput:
-                        _state[_field_name] = self.get_step_item(current_step).get_item(int(last_input))
+                        _state[_field_name] = self.get_step_item(current_step).get_item(last_input)
                     else:
                         _state[_field_name] = last_input
                 else:
                     logger.warning(
-                        f'field_name "{_field_name}" is not valid. It should be contain letters, underscores and numbers, but begin with a letter or underscore')
+                        f'field_name "{_field_name}" is not valid. It should be contain letters, underscores and '
+                        f'numbers, but begin with a letter or underscore')
                 # end variable
 
+            # call post_validation..
+
+            if current_step != 0 and last_input not in [back_symbol, home_symbol]:
+                q = self.form_questions.get(str(current_step))
+                if q:
+                    post_call = q.get('post_call')
+                    if post_call:
+                        data = {}
+                        for key in self.gather_form_keys():
+                            data[key] = r.hget(f'{msisdn}:{session_id}', key)
+                        data[self.form_questions[str(current_step)]['name']] = last_input
+
+                        f = post_call(msisdn, session_id, ussd_string, data)
             try:
 
                 resp = self.form_questions[str(current_step + 1)].copy()
@@ -183,31 +233,42 @@ class FormFlow:
             _menu = self.form_questions[str(current_step)]['menu']
             if isinstance(_menu, ListInput):
                 resp = self.invalid_input.format(menu=_menu.get_items()[4:])
+            elif callable(_menu):
+                resp = self.invalid_input.format(menu=_menu(
+                    msisdn=msisdn, session_id=session_id, ussd_string=ussd_string, data={})[4:])
             else:
                 resp = self.invalid_input.format(menu=_menu[4:])
 
-        if isinstance(resp, ListInput):
-            resp = resp.get_items()
+        # start get the response for next menu
+        if isinstance(resp['menu'], ListInput):
+            resp = resp['menu'].get_items()
 
-        elif callable(resp):
-            resp = resp()
+        elif callable(resp['menu']):
+            data = {}
+            for key in self.gather_form_keys():
+                data[key] = r.hget(f'{msisdn}:{session_id}', key)
+            if current_step != 0:
+                data[self.form_questions[str(current_step + 1)]['name']] = last_input
+
+            try:
+                resp = resp['menu'](msisdn=msisdn, session_id=session_id, ussd_string=ussd_string, data=data)
+            except TypeError as t:
+                logger.warning(t)
+                raise ImproperlyConfigured(
+                    f'The callable{resp["menu"]} should accept at least 4 arguments or accept arbitrary kwargs')
+
+        elif isinstance(resp['menu'], str):
+            resp = resp['menu']
 
-        if isinstance(resp, dict) and 'menu' in resp.keys():
-            if isinstance(resp['menu'], ListInput):
-                resp['menu'] = resp['menu'].get_items()
         return resp, _state, valid_last_input
 
-    def get_response(self, current_step, last_input):
+    def get_response(self, current_step, last_input, msisdn, session_id, ussd_string):
         if current_step is None:
-            current_step = 0
+            current_step = 1
 
-        _resp, state, valid = self._response(current_step, last_input)
-        if isinstance(_resp, dict) and 'menu' in _resp.keys():
-            message = _resp['menu']
-        else:
-            message = _resp
+        _resp, state, valid = self._response(current_step, last_input, msisdn, session_id, ussd_string)
 
-        return message, state, valid
+        return _resp, state, valid
 
 
 class NavigationMenu(Node, NodeMixin):
@@ -228,7 +289,7 @@ class NavigationMenu(Node, NodeMixin):
 
         self.all_ids = next(self._ids)
 
-    def _generate_menu(self, last_input, step=None):
+    def _generate_menu(self, last_input, msisdn, session_id, ussd_string, step=None):
         if len(self.children) == 0 and self.next_form is not None:
             # form variable is set but it is not a FormFlow class
             if not isinstance(self.next_form, FormFlow):
@@ -237,7 +298,8 @@ class NavigationMenu(Node, NodeMixin):
 
             # Here means this Node has no children but has next_form set
 
-            _message, _state, valid = getattr(self.next_form, 'get_response')(step, last_input)
+            _message, _state, valid = getattr(
+                self.next_form, 'get_response')(step, last_input, msisdn, session_id, ussd_string)
             self.menu_string = _message
             self.form_state = _state
             self.valid_last_input = valid
@@ -256,11 +318,11 @@ class NavigationMenu(Node, NodeMixin):
             self.menu_string = f"CON Select {self.title}:\n" + "\n".join(
                 [f"{i.id}. {i.title}" for i in self.children]) if self.children else ""
 
-        # if self.show_title:
-        #     self.menu_string = f'{self.title}\n{self.menu_string[4:] if self.menu_string[0:2] in ["CON", "END"] else self.menu_string}'
+        # if self.show_title: self.menu_string = f'{self.title}\n{self.menu_string[4:] if self.menu_string[0:2] in [
+        # "CON", "END"] else self.menu_string}'
 
-    def get_menu(self, last_input, step=None):
-        self._generate_menu(last_input, step)
+    def get_menu(self, last_input, msisdn, session_id, ussd_string, step=None):
+        self._generate_menu(last_input, msisdn, session_id, ussd_string, step)
         return self.menu_string, self.form_state, self.valid_last_input
 
     def _generate_id(self):
@@ -275,13 +337,19 @@ class NavigationController(BaseUSSD):
         super().__init__(msisdn, session_id, ussd_string)
         self.home_menu = home_menu
 
-    def path_processor(self, path_as_list: list = None, index=1, ):
+    def _path_process(self, path_as_list: list = None, index=1):
+        """
+
+                :param path_as_list:
+                :param index:
+                :return:
+                """
+
         path = path_as_list
         if path is None:
             path = json.loads(r.hget(self.redis_key, 'PATH_AS_LIST'))
 
         if path and path[0] in [str(back_symbol), str(home_symbol)]:
-            # invalid, we can't start by going back
             return []
 
         if len(path) == 1:
@@ -297,16 +365,16 @@ class NavigationController(BaseUSSD):
             path.pop(index - 1)
 
             if len(path) > index - 1:
-                return self.path_processor(path, index - 1)
+                return self._path_process(path, index - 1)
 
         elif path[index] == str(home_symbol):
             for i in range(index + 1):
                 path.pop(0)
 
             if len(path) > 1:
-                return self.path_processor(path, index=1)
+                return self._path_process(path, index=1)
         else:
-            return self.path_processor(path, index + 1)
+            return self._path_process(path, index + 1)
 
         return path
 
@@ -321,6 +389,16 @@ class NavigationController(BaseUSSD):
             start = start.parent
             return self.path_to_list(start, path)
 
+    def path_processor(self, path_as_list: list = None, index=1, offset: int = None):
+        if offset is None:
+            offset = 0
+
+        processed_path = self._path_process(path_as_list, index)
+
+        if len(processed_path) >= offset:
+            return processed_path[offset:]
+        return processed_path
+
     def path_navigator(self, start: NavigationMenu, path: list):
         if len(path) == 0:
             return start
@@ -359,35 +437,44 @@ class NavigationController(BaseUSSD):
                 else:
                     logger.warning(f"cannot save data of type {state[key].__class__.__name__} to redis")
 
-    def navigate(self):
+    def navigate(self, offset=None):
         step = r.hget(self.redis_key, 'FORM_STEP')
         step = int(step) if step is not None else 0
         last_input = self.ussd_string.split("*")[-1]
 
-        processed_path = self.get_processed_path()
+        # processed_path = self.get_processed_path()
+        processed_path = self.ussd_string.split("*") if self.ussd_string else []
 
         # append current input to processed_path
         # NOTE: when processed_path will be passed through path_navigator function, it will be sanitized to
         # point to the menu
 
-        if last_input:
-            processed_path.append(last_input)
+        # if last_input:
+        #     processed_path.append(last_input)
 
-        def _menu(path, add_last_input=True):
-            pro_path = self.path_processor(path.copy())
+        def _menu(path, add_last_input=True, offset=None):
+            pro_path = self.path_processor(path.copy(), offset=offset)
+            logger.info(f"PROCESSED_PATH: {pro_path}")
             _menu_ref = self.path_navigator(self.home_menu, pro_path.copy())
             # path = self.path_to_list(_menu_ref)
             # logger.info(f"MENU REF: {_menu_ref}  ::  PATH: {path}")
             r.hset(self.redis_key, 'PROCESSED_PATH', json.dumps(pro_path))
-            _resp, _state, valid_input, = getattr(_menu_ref, 'get_menu')(last_input if add_last_input else None,
-                                                                         step=step)
+
+            _resp, _state, valid_input, = getattr(_menu_ref, 'get_menu')(
+                last_input if add_last_input else None,
+                self.msisdn,
+                self.session_id,
+                self.ussd_string,
+                step=step
+            )
+
             if valid_input is not None and not valid_input:
                 r.hset(self.redis_key, 'PROCESSED_PATH', json.dumps(pro_path[:-1]))
             self._redis_processing(_state)
             return _resp
 
         try:
-            resp = _menu(processed_path)
+            resp = _menu(processed_path, offset=offset)
             r.hset(self.redis_key, 'LAST_SUCCESS_RESPONSE', resp)
         except FormBackError:
             # we pop the last path since it was pointing to a form, and now we can't go back further in the form
@@ -398,14 +485,14 @@ class NavigationController(BaseUSSD):
             processed_path.pop()
             self._redis_processing({'FORM_STEP': None})
             r.hset(self.redis_key, 'PROCESSED_PATH', json.dumps(processed_path))
-            resp = _menu(processed_path, add_last_input=False)
+            resp = _menu(processed_path, add_last_input=False, offset=offset)
             r.hset(self.redis_key, 'LAST_SUCCESS_RESPONSE', resp)
 
         except NavigationBackError:
             # we are going back inside navigation
             processed_path = self.get_processed_path()
             r.hset(self.redis_key, 'PROCESSED_PATH', json.dumps(processed_path))
-            resp = _menu(processed_path, add_last_input=False)
+            resp = _menu(processed_path, add_last_input=False, offset=offset)
             r.hset(self.redis_key, 'LAST_SUCCESS_RESPONSE', resp)
         except NavigationInvalidChoice:
             last_resp = r.hget(self.redis_key, "LAST_SUCCESS_RESPONSE")
